import uvicorn
import mlx.core as mx
from fastapi import FastAPI, Request
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse, StreamingResponse
from .engine import ModelEngine, ModelConfig
from .utils import PACKAGE_NAME, FINISH_REASON
from datetime import datetime as dt
import asyncio
import argparse
import logging, json, uuid, warnings, yaml
from typing import Union, Dict, Any, Iterator, Literal, List, Tuple

warnings.filterwarnings('ignore')
logging.basicConfig(format='[(%(levelname)s) %(asctime)s]: %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
logger = logging.getLogger(__name__)

# Configure
def get_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=f'{PACKAGE_NAME}.server',
        description='Run an OpenAI-compatible LLM server.',
    )
    parser.add_argument('-m', '--model-path', type=str, 
        default=None, help='Path to the model or the HuggingFace repository name if only one model should be served.')
    parser.add_argument('--tokenizer-path', type=str, 
        default=None, help='Path to the tokenizer or the HuggingFace repository name if only one model should be served. If None is given, it will be the model_path. Defaults to None.')
    parser.add_argument('--adapter-path', type=str, 
        default=None, help='Path to the adapter for the model. Defaults to None.')
    parser.add_argument('--revision', type=str, 
        default=None, help='Rivision of the repository if an HF repository is given. Defaults to None.')
    parser.add_argument('-q', '--quantize', type=str, 
        default='fp16', help='Model qunatization, options are "fp16", "q8", "q4", "q2". Defaults to "fp16", meaning no quantization.')
    parser.add_argument('--model-name', type=str,
        default=None, help='Model name appears in the API endpoint. If None is given, it will be created automatically with the model path. Defaults to None.')
    parser.add_argument('-cf', '--config-file', type=str, 
        default=None, 
        help='Path of the config file that store the configs of all models wanted to be served. If this is passed, "model-path", "quantize", and "model-name" will be ignored.')
    parser.add_argument('--prefill-step-size', type=int, 
        default=512, help='Batch size for model prompt processing. Defaults to 512.')
    parser.add_argument('-mk', '--max-keep', type=int, 
        default=10, help='Maximum number of cache history for each model to keep. Defaults to 10.')
    parser.add_argument('--token-threshold', type=int,
        default=50, 
        help='Minimum number of tokens in the prompt plus generated text to trigger prompt caching. Shorter prompts do not require caching to speed up generation. Defaults to 50.')
    parser.add_argument('--api-key', type=str, default=None, help='API key to access the endpoints. Defaults to None.')
    parser.add_argument('-p', '--port', type=int, 
                        default=5001, help='Port to server the API endpoints.')
    return parser

def parse_args() -> Tuple[ModelEngine, int, List[str]]:
    engine_args = get_arg_parser().parse_args()
    port = engine_args.port
    api_keys = [engine_args.api_key] if engine_args.api_key is not None else []
    if engine_args.config_file is not None:
        with open(engine_args.config_file, 'r') as f:
            model_args = yaml.safe_load(f)
        if isinstance(model_args, list):
            model_args = [ModelConfig(**args) for args in model_args]
        else:
            model_args = [ModelConfig(**model_args)]
    elif engine_args.model_path is not None:
        model_args = [
            ModelConfig(
                model_id_or_path=engine_args.model_path,
                tokenizer_id_or_path=engine_args.tokenizer_path,
                adapter_path=engine_args.adapter_path,
                quant=engine_args.quantize,
                revision=engine_args.revision,
                model_name=engine_args.model_name
            )
        ]
    else:
        raise ValueError('Either model_path or config_file has to be provide.')
    engine = ModelEngine(models=model_args, prefill_step_size=engine_args.prefill_step_size, 
        token_threshold=engine_args.token_threshold, max_keep=engine_args.max_keep, logger=logger)
    return engine, port, api_keys


engine, port, api_keys = parse_args()

models = [dict(id=m, object='model', created=int(dt.now().timestamp()), owned_by='null', permission=[], root='root') for m in engine.models.keys()]
app = FastAPI()
semaphore = asyncio.Semaphore(1)

def get_return_dict(
        text: Union[str, List[str]], 
        id: str, model: str, 
        return_type: Literal['text_completion', 'chat.completion'], 
        finish_reason: FINISH_REASON,
        created: float,
        stream: bool,
        prompt_tokens: int = 100,
        total_tokens: int = 200,
        logprob_dict: Dict[str, Any] = None
    ) -> Dict[str, Any]:
    if return_type == 'text_completion':
        choices = [dict(index=0, finish_reason=finish_reason, text=t) for t in text] if isinstance(text, list) else [dict(index=0, finish_reason=finish_reason, text=text, logprobs=logprob_dict)]
        return dict(id='cmpl-'+id, 
            object=return_type, 
            created=created,
            model=model, 
            choices=choices,
            usage=dict(
                prompt_tokens=prompt_tokens, 
                completion_tokens=total_tokens - prompt_tokens, 
                total_tokens=total_tokens
            )
        )
    elif ((return_type == 'chat.completion') and (stream)):
        return dict(
            id='chatcmpl-'+id, 
            object=return_type + '.chunk', 
            created=created, 
            model=model, 
            choices=[
                dict(
                    index=0, 
                    finish_reason=finish_reason, 
                    delta=dict(
                        role="assistant",
                        content=text
                    ),
                    logprobs=logprob_dict
                )
            ]
        )
    elif return_type == 'chat.completion':
        return dict(
            id='chatcmpl-'+id, 
            object=return_type, 
            created=created, 
            model=model, 
            usage=dict(
                prompt_tokens=prompt_tokens, 
                completion_tokens=total_tokens - prompt_tokens, 
                total_tokens=total_tokens
            ), 
            choices=[
                dict(
                    index=0, 
                    message=dict(
                        role="assistant", 
                        content=text[0], 
                        tool_calls=[]
                    ), 
                    logprobs=logprob_dict,
                    finish_reason=finish_reason
                )
            ]
        )
    
def create_logprobs(token: int, logprobs: mx.array, top: int) -> Tuple[float, Dict[int, float]]:
    sorted_indices = mx.argpartition(-logprobs, kth=top - 1)
    top_indices = sorted_indices[: top]
    top_logprobs = logprobs[top_indices]
    top_token_info = zip(top_indices.tolist(), top_logprobs.tolist())
    return logprobs[token].item(), dict(top_token_info)

# Generation
def stream_generate(args: Dict[str, Any]) -> Iterator[str]:
    extra_body = args.get('extra_body', dict())
    id = uuid.uuid4().hex
    response = engine.generate(
        model_name=args['model'],
        prompt=args['prompt'],
        stop=args.get('stop'),
        max_new_tokens=args.get('max_tokens', 1024),
        temperature=args.get('temperature', 0.0),
        top_p=args.get('top_p', 1),
        repetition_penalty=args.get('frequency_penalty', None),
        seed=args.get('seed'),
        stream=True,
        **extra_body
    )
    created = int(dt.now().timestamp())
    logprobs = args.get('logprobs', False)
    top_logprobs = args.get('top_logprobs', 3)
    token_count = 0
    for i in response:
        token_count += 1
        if logprobs:
            token_prob, top_dict = create_logprobs(i.token, i.logprobs, top_logprobs)
            logprobs_dict = dict(
                content=[
                        dict(
                            token=engine.tokenizer.decode(i.token),
                            logprob=token_prob,
                            top_logprobs=[
                                dict(
                                    token=engine.tokenizer.decode(k),
                                    logprob=v
                                )
                                for k, v in top_dict.items()
                            ]
                        )
                    ]
                )
        else:
            logprobs_dict = None
        return_dict = get_return_dict(text=i.text, id=id, 
            model=args['model'], return_type=args['endpoint'], 
            stream=True, finish_reason=i.finish_reason, created=created, prompt_tokens=i.prompt_len, total_tokens=token_count + i.prompt_len,
            logprob_dict=logprobs_dict)
        yield f'data: {json.dumps(return_dict)}\n\n'
    yield 'data: [DONE]'

def static_generate(args: Dict[str, Any]) -> Dict[str, Any]:
    extra_body = args.get('extra_body', dict())
    id = uuid.uuid4().hex
    created = int(dt.now().timestamp())
    if isinstance(args['prompt'], str):
        response = engine.generate(
            model_name=args['model'],
            prompt=args['prompt'],
            stop=args.get('stop'),
            max_new_tokens=args.get('max_tokens', 1024),
            temperature=args.get('temperature', 0.0),
            top_p=args.get('top_p', 1),
            repetition_penalty=args.get('frequency_penalty', None),
            seed=args.get('seed'),
            stream=False,
            **extra_body
        )
        outputs = [response]
    else:
        outputs = []
        for prompt in args['prompt']:
            response = engine.generate(
                model_name=args['model'],
                prompt=prompt,
                stop=args.get('stop'),
                max_new_tokens=args.get('max_tokens', 1024),
                temperature=args.get('temperature', 0.0),
                top_p=args.get('top_p', 1),
                repetition_penalty=args.get('frequency_penalty', None),
                stream=False,
                **extra_body
            )
            outputs.append(response)
    texts = list(map(lambda x: x.text, outputs))
    prompt_len = sum(list(map(lambda x: x.prompt_len, outputs)))
    total_tokens = sum(list(map(lambda x: len(x.token_ids), outputs)))
    return_dict = get_return_dict(text=texts, id=id, model=args['model'], return_type=args['endpoint'], 
            stream=False, created=created, finish_reason=outputs[-1].finish_reason, prompt_tokens=prompt_len, total_tokens=total_tokens)
    return return_dict
    
async def async_generate_stream(args: Dict[str, Any]):
    generator = await asyncio.to_thread(stream_generate, args)
    for token in generator:
        yield token

async def async_generate(args: Dict[str, Any]):
    result = await asyncio.to_thread(static_generate, args)
    return result

async def async_chat_generate_stream(args: Dict[str, Any]):
    engine._switch_model(args['model'])
    if engine.tokenizer.chat_template is None:
            engine.tokenizer.chat_template = engine.tokenizer.default_chat_template
    prompt = engine.tokenizer.apply_chat_template(
            args['messages'], tokenize=True, add_generation_prompt=True
        )
    if ((len(prompt) > 0) and (prompt[0] == engine.tokenizer.bos_token_id)):
        prompt = prompt[1:]
    args['prompt'] = engine.tokenizer.decode(prompt, skip_special_tokens=False)
    generator = await asyncio.to_thread(stream_generate, args)
    for token in generator:
        yield token

async def async_chat_generate(args: Dict[str, Any]):
    engine._switch_model(args['model'])
    if engine.tokenizer.chat_template is None:
            engine.tokenizer.chat_template = engine.tokenizer.default_chat_template
    prompt = engine.tokenizer.apply_chat_template(
            args['messages'], tokenize=True, add_generation_prompt=True
        )
    if ((len(prompt) > 0) and (prompt[0] == engine.tokenizer.bos_token_id)):
        prompt = prompt[1:]
    args['prompt'] = engine.tokenizer.decode(prompt, skip_special_tokens=False)
    result = await asyncio.to_thread(static_generate, args)
    return result

@app.post('/v1/completions',  response_model=None)
async def completions(request: Request) -> Union[StreamingResponse, JSONResponse]:
    content = await request.json()
    content['endpoint'] = 'text_completion'
    api_key = request.headers.get('authorization', 'Bearer ').removeprefix('Bearer ')
    if ((api_key not in api_keys) and (len(api_keys) != 0)):
        return JSONResponse(jsonable_encoder(dict(error='Invalid API key.')), status_code=404)
    model = content.get('model')
    if model not in engine.models.keys():
        return JSONResponse(jsonable_encoder(dict(error=f'Model "{model}" does not exist.')), status_code=404)
    stream = content.get('stream', False)
    async with semaphore:
        if ((stream) and (isinstance(content['prompt'], str))):
            return StreamingResponse(async_generate_stream(content), media_type="text/event-stream")
        else:
            result = await async_generate(content)
            return JSONResponse(jsonable_encoder(result))
        
@app.post('/v1/chat/completions',  response_model=None)
async def completions(request: Request) -> Union[StreamingResponse, JSONResponse]:
    content = await request.json()
    content['endpoint'] = 'chat.completion'
    api_key = request.headers.get('authorization', 'Bearer ').removeprefix('Bearer ')
    if ((api_key not in api_keys) and (len(api_keys) != 0)):
        return JSONResponse(jsonable_encoder(dict(error='Invalid API key.')), status_code=404)
    model = content.get('model')
    if model not in engine.models.keys():
        return JSONResponse(jsonable_encoder(dict(error=f'Model "{model}" does not exist.')), status_code=404)
    stream = content.get('stream', False)
    async with semaphore:
        if stream:
            return StreamingResponse(async_chat_generate_stream(content), media_type="text/event-stream")
        else:
            result = await async_chat_generate(content)
            return JSONResponse(jsonable_encoder(result))

@app.get('/v1/models')
async def get_models(request: Request) -> JSONResponse:
    api_key = request.headers.get('authorization', 'Bearer ').removeprefix('Bearer ')
    if ((api_key not in api_keys) and (len(api_keys) != 0)):
        return JSONResponse(jsonable_encoder(dict(error='Invalid API key.')), status_code=404)
    return JSONResponse(content=jsonable_encoder(dict(object='list', data=models))) 

def main():
    uvicorn.run(app, port=port)

if __name__ == '__main__':
    main()



from mlx_lm.utils import load, get_model_path, convert, load_config, load_model
from mlx_lm.tokenizer_utils import load_tokenizer
from mlx_lm.tokenizer_utils import TokenizerWrapper
from mlx.nn import Module
import mlx.core as mx
from .utils import mlx_cache_dir, generate, stream_generate, remove_bos_duplicates, FINISH_REASON
from .caching import CacheManager
from pathlib import Path
from logging import Logger
import os
from typing import NamedTuple, Optional, List, Dict, Union, Literal, Any, Tuple, Iterator

class ModelConfig(NamedTuple):
    model_id_or_path: str
    tokenizer_id_or_path: Optional[str] = None
    adapter_path: Optional[str] = None
    quant: Literal['fp16', 'q8', 'q4', 'q2'] = 'fp16'
    revision: Optional[str] = None
    model_name: Optional[str] = None
    model_config: Optional[Dict[str, Any]] = None
    tokenizer_config: Optional[Dict[str, Any]] = None

def get_model_name(model_id_or_path: str) -> str:
    """Getting a clean name for the model.

    Args:
        model_id_or_path (str): HuggingFace repository name or the local path of the model.

    Returns:
        str: A clean name for the model.
    """
    base_name = model_id_or_path.split('/')[-1]
    base_name = base_name.split('\\')[-1]
    base_name = base_name.lower().replace('_', '-').strip()
    return base_name

def make_model_exist(
        model_id_or_path: str,
        quant: Literal['fp16', 'q8', 'q4', 'q2'] = 'fp16',
        revision: Optional[str] = None,
        adapter_path: Optional[str] = None
    ) -> Path:
    """Check if the the model exists locally, if not, it will attempt to download the model and convert it to the appropriate quant.

    Args:
        model_id_or_path (str): HuggingFace repository name or the local path of the model.
        quant (Literal[&#39;fp16&#39;, &#39;q8&#39;, &#39;q4&#39;, &#39;q2&#39;], optional): Quantisation of the model. Defaults to 'fp16'.
        revision (Optional[str], optional): Revision of the repository if a HuggingFace repository name is given. Defaults to None.
        adapter_path (Optional[str], optional): Check the existence of the adapter file if given. Defaults to None.

    Returns:
        Path: Path to the local model to serve.
    """
    model_base_name = get_model_name(model_id_or_path=model_id_or_path)
    if quant != 'fp16':
        model_base_name += f'-{quant}'
    mlx_path = os.path.join(mlx_cache_dir(), model_base_name)
    
    if os.path.exists(mlx_path):
        if adapter_path is not None:
            if not os.path.exists(adapter_path):
                raise FileNotFoundError(f'Adapter not found in "{adapter_path}".')
        return Path(mlx_path)
    else:
        hf_path = get_model_path(path_or_hf_repo=model_id_or_path, revision=revision)
        if quant == 'fp16':
            mlx_path = hf_path
        else:
            config = load_config(model_path=hf_path)
            quantisation = config.get('quantization', {}).get('bits')
            quant_target = int(quant.removeprefix('q'))
            if quant_target == quantisation:
                mlx_path = hf_path
            else:
                model_base_name = get_model_name(model_id_or_path=model_id_or_path)
                mlx_path = os.path.join(mlx_cache_dir(), model_base_name + f'-q{quant_target}')
                if not os.path.exists(mlx_path):
                    convert(hf_path=hf_path, mlx_path=mlx_path, quantize=True, q_bits=int(quant[-1]))
        if adapter_path is not None:
            if not os.path.exists(adapter_path):
                raise FileNotFoundError(f'Adapter not found in "{adapter_path}".')
        return Path(mlx_path)

def get_model_and_tokenizer(
        model_id_or_path: str,
        tokenizer_id_or_path: Optional[str] = None,
        quant: Literal['fp16', 'q8', 'q4', 'q2'] = 'fp16',
        revision: Optional[str] = None,
        adapter_path: Optional[str] = None,
        model_config: Optional[Dict[str, Any]] = None,
        tokenizer_config: Optional[Dict[str, Any]] = None
    ) -> Tuple[Module, TokenizerWrapper]:
    """Load the llm model and it's tokenizer.

    Args:
        model_id_or_path (str): HuggingFace repository name or the local path of the model.
        tokenizer_id_or_path (Optional[str], optional): HuggingFace repository name or the local path of the model. If None is given, `model_id_or_path` will be used. Defaults to None.
        quant (Literal[&#39;fp16&#39;, &#39;q8&#39;, &#39;q4&#39;, &#39;q2&#39;], optional): Quantisation of the model. Defaults to 'fp16'.
        revision (Optional[str], optional): Revision of the repository if a HuggingFace repository name is given. Defaults to None.
        adapter_path (Optional[str], optional): The adapter file for the model. Defaults to None.
        model_config (Optional[Dict[str, Any]], optional): Extra arguments for loading the model. Defaults to None.
        tokenizer_config (Optional[Dict[str, Any]], optional): Extra arguments for loading the tokenizer. Defaults to None.

    Returns:
        Tuple[Module, TokenizerWrapper]: The model and the tokenizer.
    """
    mlx_path = make_model_exist(model_id_or_path=model_id_or_path, quant=quant, revision=revision, adapter_path=adapter_path)
    tokenizer_config = {} if tokenizer_config is None else tokenizer_config
    model_config = {} if model_config is None else model_config
    model = load_model(model_path=mlx_path, model_config=model_config)
    tokenizer = load_tokenizer(model_path=mlx_path if tokenizer_id_or_path is None else Path(tokenizer_id_or_path), tokenizer_config_extra=tokenizer_config)
    return model, tokenizer

class EngineOutput(NamedTuple):
    text: str
    token: int
    token_ids: List[int]
    logprobs: mx.array
    prompt_len: int
    finish_reason: FINISH_REASON

class ModelEngine:

    def __init__(self, 
            models: Union[ModelConfig, List[ModelConfig]],
            prefill_step_size: int = 512,
            token_threshold: int = 50, 
            max_keep: int = 10, 
            logger: Optional[Logger] = None
        ) -> None:
        """Initialising the engine.

        Args:
            models (Union[ModelConfig, List[ModelConfig]]): Model configurations or list of model configurations to serve.
            prefill_step_size (int, optional): Batch size for prompt preprocessing. Defaults to 512.
            token_threshold (int, optional): Minimum number of tokens to be considered to save as a prompt cache history. Defaults to 50.
            max_keep (int, optional): Maximum number of cache history files to keep. Defaults to 10.
            logger (Optional[Logger], optional): If a logger is provided, messages of adding and deleting caches will be added.
        """
        self.model_configs = [models] if isinstance(models, ModelConfig) else models
        self.prefill_step_size = prefill_step_size
        self.token_threshold = token_threshold
        self.max_keep = max_keep
        self.logger = logger
        self.models: Dict[str, ModelConfig] = dict()
        self.model: Optional[Module] = None
        self.tokenizer: Optional[TokenizerWrapper] = None
        self.current_model: Optional[str] = None
        self.cache_manager: Optional[CacheManager] = None
        self.cache_dir = os.path.join(mlx_cache_dir(), 'prompt_cache')
        for model_config in self.model_configs:
            self._prepare_model(model_config=model_config)

    def _prepare_model(self, model_config: ModelConfig) -> None:
        """Making sure the given model is available locally.

        Args:
            model_config (ModelConfig): Model configuaration of the model to check.
        """
        model_path = model_config.model_id_or_path
        adapter_path = model_config.adapter_path
        quant = model_config.quant
        model_name = model_config.model_name
        if model_name is None:
            model_name = get_model_name(model_path)
            if quant != 'fp16':
                model_name += f'-{quant}'
            if adapter_path:
                model_name += f'-{get_model_name(adapter_path)}'
        if model_name in self.models.keys():
            if self.logger:
                self.logger.error(f'Model "{model_name}" already exists. If there are different configurations with the same model name, please use a different model name.')
                return
            else:
                raise ValueError(f'Model "{model_name}" already exists. If there are different configurations with the same model name, please use a different model name.')
        mlx_path = make_model_exist(model_id_or_path=model_path, quant=quant, revision=model_config.revision, adapter_path=adapter_path)
        mx.metal.clear_cache()
        self.models[model_name] = model_config

    def _switch_model(self, model_name: str) -> None:
        """Switch the model in ram to the given model.

        Args:
            model_name (str): Model name of the model.
        """
        if model_name not in self.models.keys():
            raise ValueError(f'No model named "{model_name}".')
        if model_name != self.current_model:
            del self.model, self.tokenizer, self.cache_manager
            mx.metal.clear_cache()
            model_args = self.models[model_name]._asdict()
            model_args.pop('model_name')
            self.model, self.tokenizer = get_model_and_tokenizer(**model_args)
            self.current_model = model_name
            self.cache_manager = CacheManager(
                cache_dir=os.path.join(self.cache_dir, model_name),
                token_threshold=self.token_threshold,
                max_keep=self.max_keep,
                logger=self.logger
                )
    
    def generate(self,
            model_name: str,
            prompt: str,
            stream: bool = False,
            stop: Optional[List[str]] = None,
            max_new_tokens: int = 256,
            temperature: float = 0,
            repetition_penalty: Optional[float] = None,
            repetition_context_size: Optional[int] = 100,
            top_p: float = 1.0,
            min_p: float = 0.0,
            min_tokens_to_keep: int = 1,
            logit_bias: Optional[Dict[int, float]] = None,
            prefill_step_size: Optional[int] = None,
            seed: Optional[int] = None,
            **kwargs
        ) -> Union[EngineOutput, Iterator[EngineOutput]]:
        """Generate text with a model.

        Args:
            model_name (str): Name of the model to use.
            prompt (str): Text prompt for the generation.
            stream (bool, optional): Whether to return a generator for streaming text tokens during generation. Defaults to False.
            stop (Optional[List[str]], optional): List of texts to stop generation. Defaults to None.
            max_new_tokens (int, optional): Maximum number of tokens to generate. Defaults to 256.
            temperature (float, optional): Temperature that decide the "randomness" of the generation. Defaults to 0.
            repetition_penalty (Optional[float], optional): Penalty for repeated textt during generation. 1.0 being no penalty while higher values gives more penalty. Defaults to None.
            repetition_context_size (Optional[int], optional): Size of previous tokens to consider for repetition penalty. Defaults to 100.
            top_p (float, optional): Top P sampling. Defaults to 1.0.
            min_p (float, optional): Min P sampling. Defaults to 0.0.
            min_tokens_to_keep (int, optional): Minimum number of tokens to keep for sampling. Defaults to 1.
            logit_bias (Optional[Dict[int, float]], optional): Logits bias. Defaults to None.
            prefill_step_size (Optional[int], optional): Batch size for prompt preprocessing. Defaults to None.

        Returns:
            Union[str, Iterator[str]]: Generated text if `stream=False`.

        Yields:
            Iterator[Union[str, Iterator[str]]]: Generator of generated text if `stream=True`.
        """
        self._switch_model(model_name=model_name)
        prompt_tokens = self.tokenizer.encode(prompt)
        prompt_tokens = remove_bos_duplicates(prompt_tokens, self.tokenizer.bos_token_id)
        cache, cache_id = self.cache_manager.find_cache(token_ids=prompt_tokens)
        gen_kwargs = dict(
            max_tokens=max_new_tokens,
            stop=stop,
            temp=temperature,
            repetition_penalty=repetition_penalty,
            repetition_context_size=repetition_context_size,
            top_p=top_p,
            min_p=min_p,
            min_tokens_to_keep=min_tokens_to_keep,
            logit_bias=logit_bias,
            prefill_step_size=self.prefill_step_size if prefill_step_size is None else prefill_step_size,
            cache_history=cache,
            seed=seed,
            verbose=True
        )
        if stream:
            def stream_output():
                nonlocal cache
                try:
                    for output in stream_generate(
                        model=self.model,
                        tokenizer=self.tokenizer,
                        prompt=prompt,
                        **gen_kwargs
                    ):
                        cache = output.step
                        yield EngineOutput(
                            text=output.text,
                            token=output.step.token,
                            token_ids=output.step.token_ids,
                            logprobs=output.step.logprobs,
                            prompt_len=output.prompt_len,
                            finish_reason=output.finish_reason
                        )
                finally:
                    self.cache_manager.save_cache(cache=cache, from_cache_id=cache_id)
                    del cache, output
                    mx.metal.clear_cache()

            return stream_output()
        else:
            output = generate(
                model=self.model,
                tokenizer=self.tokenizer,
                prompt=prompt,
                **gen_kwargs
            )
            cache = output.step
            engine_output = EngineOutput(
                    text=output.text,
                    token=output.step.token,
                    token_ids=output.step.token_ids,
                    logprobs=output.step.logprobs,
                    prompt_len=output.prompt_len,
                    finish_reason=output.finish_reason
                )
            self.cache_manager.save_cache(cache=cache, from_cache_id=cache_id)
            del cache, output
            mx.metal.clear_cache()
            return engine_output



from mlx_lm.utils import make_kv_caches, load, convert, apply_repetition_penalty
from mlx_lm.models.base import KVCache, RotatingKVCache
from mlx_lm.sample_utils import categorical_sampling, min_p_sampling, top_p_sampling
from mlx_lm.tokenizer_utils import TokenizerWrapper
import mlx.core as mx
import mlx.nn as nn
import time, os
from transformers import PreTrainedTokenizer
from typing import Union, Optional, List, Tuple, Dict, Generator, NamedTuple, Callable, Iterator, Literal
os.environ['TOKENIZERS_PARALLELISM'] = 'true'

PACKAGE_NAME = 'mlx_textgen'

FINISH_REASON = Optional[Literal['stop', 'length']]

def mlx_cache_dir() -> str:
    home_dir = os.path.expanduser('~')
    mlx_cache_dir = os.path.join(home_dir, '.cache', PACKAGE_NAME)
    os.makedirs(mlx_cache_dir, exist_ok=True)
    return mlx_cache_dir

class StepOutput(NamedTuple):
    token: int
    logprobs: mx.array
    token_ids: List[int]
    cache: Optional[List[Union[KVCache, RotatingKVCache]]] = None

class CacheHistory(NamedTuple):
    cache: List[Tuple[mx.array, mx.array]]
    token_ids: List[int]

class GenerateOutput(NamedTuple):
    text: str
    step: StepOutput
    finish_reason: FINISH_REASON
    prompt_len: int

class StopCondition(NamedTuple):
    stop_met: bool
    trim_length: int

def convert_cache_to_history(cache: StepOutput) -> CacheHistory:
    """Helper function to convert the output of "generate_step" into reusable cache history.

    Args:
        cache (StepOutput): Outout of "generate_step".

    Returns:
        CacheHistory: Reusable cache history.
    """
    cache_list = [(c.state[0][..., : c.offset, :], c.state[1][..., : c.offset, :]) for c in cache.cache]
    return CacheHistory(cache=cache_list, token_ids=cache.token_ids)

def save_cache(cache: Union[StepOutput, CacheHistory], filename: str, metadata: Optional[Dict[str, str]] = None) -> None:
    """Saving a prompt cache into a disk.

    Args:
        cache (Union[StepOutput, CacheHistory]): Output of "generate_step" or formatted cache history.
        filename (str): File directory of the prompt cache safetensors file.
        metadata (Optional[Dict[str, str]], optional): String keys and values metadata for the cache file. Defaults to None.
    """
    import orjson
    cache = cache if isinstance(cache, CacheHistory) else convert_cache_to_history(cache=cache)
    metadata = dict() if not isinstance(metadata, dict) else metadata
    metadata['token_ids'] = orjson.dumps(cache.token_ids).decode()
    cache_dict = {}
    for i, c in enumerate(cache.cache):
        cache_dict[f'{i}_key'] = c[0]
        cache_dict[f'{i}_value'] = c[1]
    mx.save_safetensors(file=filename, arrays=cache_dict, metadata=metadata)
    mx.metal.clear_cache()

def load_cache(filename: str) -> Tuple[CacheHistory, Dict[str, str]]:
    """Loading prompt cache from a safetnesors file into ram for generation.

    Args:
        filename (str): File directory of the prompt cache safetensors file. 

    Returns:
        Tuple[CacheHistory, Dict[str, str]]: Reusable cache history and the metadata.
    """
    import orjson
    cache_dict, metadata = mx.load(filename, return_metadata=True)

    # Loading cache
    num_layers = int(len(cache_dict) / 2)
    cache = []
    for i in range(num_layers):
        cache.append((cache_dict[f'{i}_key'], cache_dict[f'{i}_value']))
    token_ids = orjson.loads(metadata.pop('token_ids'))
    ch = CacheHistory(cache=cache, token_ids=token_ids)
    mx.metal.clear_cache()
    return ch, metadata

def remove_bos_duplicates(token_ids: List[int], bos_token_id: Optional[int]) -> List[int]:
    """Remove duplicates of bos tokens if there are multiple of them in the token id list. This usually happen when the prompt was already formatted by the tokenizer.

    Args:
        token_ids (List[int]): List of token ids.
        bos_token_id (Optional[int]): BOS token id.

    Returns:
        List[int]: List of token ids without bos token duplicates.
    """
    if ((len(token_ids) > 1) and (token_ids[0] == bos_token_id)):
        while token_ids[1] == bos_token_id:
            token_ids = token_ids[:1] + token_ids[2:]
    return token_ids

def find_max_prefix_num(new: List[int], baseline: List[int]) -> int:
    """Helper function to find the maximum number of tokens shared in the prefix of two prompts.

    Args:
        new (List[int]): First prompt token ids.
        baseline (List[int]): Second prompt token ids.

    Returns:
        int: The maximum number of tokens shared in the prefix of two prompts.
    """
    from itertools import takewhile
    return len(list(takewhile(lambda x: x[0] == x[1], zip(new, baseline))))

def stopping_criteria(
        text: str,
        stop_tuple: List[Tuple[str, int]],
        eos_tuple: Union[Tuple[str, int], None],
    ) -> StopCondition:
    """Get the stopping condition with stop words and eos token.

    Args:
        text (str): Text to be determined to stop or not.
        stop_tuple (List[Tuple[str, int]]): List of tuple with the stop word string and the length of the string. Must be ordered descendingly by length.
        eos_tuple (Union[Tuple[str, int], None]): Eos token string and it's length.

    Returns:
        StopCondition: A status class stating whether the generation should stop and the length of text to trim.
    """
    if eos_tuple is not None and eos_tuple[0] in text:
        return StopCondition(stop_met=True, trim_length=len(text.split(eos_tuple[0])[-1]) + eos_tuple[1])

    return next(
        (StopCondition(stop_met=True, trim_length=length + len(text.split(stop)[-1]))
        for stop, length in stop_tuple if stop in text), StopCondition(stop_met=False, trim_length=0)
    )

def sequence_overlap(s1: str, s2: str) -> bool:
    """Determine if there is overlapping string between the suffix of the first string and the prefix of the second string.

    Args:
        s1 (str): First string.
        s2 (str): Second string.

    Returns:
        bool: Whether there is overlap.
    """
    max_overlap = min(len(s1), len(s2))
    return any(s1[-i:] == s2[:i] for i in range(1, max_overlap + 1))

def get_kv_caches(
        model: nn.Module, 
        promp_tokens: List[int],
        max_kv_size: Optional[int] = None, 
        cache_history: Optional[Union[CacheHistory, StepOutput]] = None
    ) -> Tuple[List[Union[KVCache, RotatingKVCache]], int]:
    """Helper function to setup the kv cache in "generate_step".

    Args:
        model (nn.Module): The LLM model.
        promp_tokens (List[int]): Prompt tokens ids.
        max_kv_size (Optional[int], optional): Maximum size of the key-value cache. Old entries (except the first 4 tokens) will be overwritten.. Defaults to None.
        cache_history (Optional[Union[CacheHistory, StepOutput]], optional): Reusable prompt cache history or previous generation step output. Defaults to None.

    Returns:
        Tuple[List[Union[KVCache, RotatingKVCache]], int]: List of KV cache for model generation and the number of tokens reused from the cache history.
    """
    cache = make_kv_caches(model=model, max_kv_size=max_kv_size)
    max_prefix = 0

    if cache_history is not None:
        if isinstance(cache_history, StepOutput):
            cache_history = convert_cache_to_history(cache=cache_history)
        if len(cache_history.cache) != len(cache):
            raise ValueError("Wrong number of layers in the cache history")
        cache_size = cache_history.cache[0][0].shape[2]
        if (max_kv_size is None) or (cache_size <= max_kv_size):
            max_prefix = find_max_prefix_num(promp_tokens, cache_history.token_ids)
            # Leave at least one token to evaluate during generation.
            max_prefix = max_prefix - 1 if len(promp_tokens) == max_prefix else max_prefix

            # Set the history in the cache objects and evaluate them to prepare for
            # generation.
            for c, h in zip(cache, cache_history.cache):
                c.update_and_fetch(h[0][:, :, :max_prefix, :], h[1][:, :, :max_prefix, :])
            mx.eval([c.state for c in cache])
    return cache, max_prefix
    
def generate_step(
    prompt: mx.array,
    model: nn.Module,
    temp: float = 0.0,
    repetition_penalty: Optional[float] = None,
    repetition_context_size: Optional[int] = 20,
    top_p: float = 1.0,
    min_p: float = 0.0,
    min_tokens_to_keep: int = 1,
    seed: Optional[int] = None,
    logit_bias: Optional[Dict[int, float]] = None,
    prefill_step_size: int = 512,
    verbose: bool = False,
    max_kv_size: Optional[int] = None,
    cache_history: Optional[Union[CacheHistory, StepOutput]] = None
) -> Generator[StepOutput, None, None]:
    """
    A generator producing token ids based on the given prompt from the model.

    Args:
        prompt (mx.array): The input prompt.
        model (nn.Module): The model to use for generation.
        temp (float): The temperature for sampling, if 0 the argmax is used.
          Default: ``0``.
        repetition_penalty (float, optional): The penalty factor for repeating
          tokens.
        repetition_context_size (int, optional): The number of tokens to
          consider for repetition penalty. Default: ``20``.
        top_p (float, optional): Nulceus sampling, higher means model considers
          more less likely words.
        min_p (float, optional): The minimum value (scaled by the top token's
          probability) that a token probability must have to be considered.
        min_tokens_to_keep (int, optional): Minimum number of tokens that cannot
          be filtered by min_p sampling.
        seed (Optional[int], optional): Random seed for sampling. Defaults to None.
        logit_bias (dictionary, optional): Additive logit bias.
        prefill_step_size (int, optional): Step size for processing the prompt. Defaults to 512.
        verbose (bool, optional): 
        max_kv_size (int, optional): Maximum size of the key-value cache. Old
          entries (except the first 4 tokens) will be overwritten.
        cache_history (Optional[Union[CacheHistory, StepOutput]], optional): Reusable prompt cache history or previous generation step output. Defaults to None.

    Yields:
        Generator[StepOutput, None, None]: A generator producing
          one token and a vector of log probabilities.
    """

    def sample(logits: mx.array) -> Tuple[mx.array, float]:
        if logit_bias:
            indices = mx.array(list(logit_bias.keys()))
            values = mx.array(list(logit_bias.values()))
            logits[:, indices] += values
        logprobs = logits - mx.logsumexp(logits)

        if temp == 0:
            token = mx.argmax(logits, axis=-1)
        else:
            if top_p > 0 and top_p < 1.0:
                token = top_p_sampling(logits, top_p, temp)
            elif min_p != 0.0:
                token = min_p_sampling(logits, min_p, min_tokens_to_keep, temp)
            else:
                token = categorical_sampling(logits, temp)
        return token, logprobs

    if repetition_penalty and (
        repetition_penalty < 0 or not isinstance(repetition_penalty, float)
    ):
        raise ValueError(
            f"repetition_penalty must be a non-negative float, got {repetition_penalty}"
        )
    tokens: List[int] = prompt.tolist()
    token_count = len(tokens)
    y = prompt

    if seed:
        mx.random.seed(seed)

    # Create the KV cache for generation and get the number of tokens being reused.
    cache, max_prefix = get_kv_caches(model=model, promp_tokens=tokens, max_kv_size=max_kv_size, cache_history=cache_history)
    y = y[max_prefix:]

    repetition_context = tokens

    if repetition_context_size:
        repetition_context = repetition_context[-repetition_context_size:]

    def _step(y):
        nonlocal repetition_context
        logits = model(y[None], cache=cache)
        logits = logits[:, -1, :]

        if repetition_penalty:
            logits = apply_repetition_penalty(
                logits, repetition_context, repetition_penalty
            )
            y, logprobs = sample(logits)
            repetition_context.append(y.item())
        else:
            y, logprobs = sample(logits)

        if repetition_context_size:
            if len(repetition_context) > repetition_context_size:
                repetition_context = repetition_context[-repetition_context_size:]
        return y, logprobs.squeeze(0)

    # Getting preprocessing batches
    num_batches = y.shape[0] // prefill_step_size
    if num_batches != (y.size / prefill_step_size):
        num_batches += 1
    batches = [(i * prefill_step_size, min((i + 1) * prefill_step_size, y.size)) for i in range(num_batches)]
    num_tokens = y.size

    # Prompt preprocessing
    if verbose:
        from tqdm import tqdm
        batches = tqdm(batches)
    pp_start = time.perf_counter()
    for b in batches:
        if verbose:
            batches.set_description(f'Processing prompt ({b[1]}/{y.size})')
        if (b[1] - b[0]) >= prefill_step_size:
            model(y[b[0]:b[1]][None], cache=cache)
            mx.eval([c.state for c in cache])
            mx.metal.clear_cache() # Clearing mlx cache, otherwise it grows very quick with longer prompts.
        else:
            y = y[b[0]:b[1]]
            y, logprobs = _step(y)
    pp_end = time.perf_counter() - pp_start
    if verbose:
        print(f'Prompt preprocessing time for {num_tokens} tokens: {pp_end:.4}s ({num_tokens/pp_end:.4f} tok/sec)')

    mx.async_eval(y)
    while True:
        next_y, next_logprobs = _step(y)
        mx.async_eval(next_y)
        token = y.item()
        tokens.append(token)
        token_count += 1
        if max_kv_size is not None:
            # Trim off token ids if max_kv_size is set.
            if token_count >= max_kv_size:
                token_count -= 1
                tokens = tokens[:4] + tokens[5:]
        yield StepOutput(token=token, logprobs=logprobs, token_ids=tokens, cache=cache)
        y, logprobs = next_y, next_logprobs

def stream_generate(
        model: nn.Module, 
        tokenizer: TokenizerWrapper, 
        prompt: str,
        max_tokens: int = 512,
        stop: Optional[List[str]] = None,
        cache_history: Union[StepOutput, CacheHistory] = None,
        verbose: bool = False,
        **kwargs
    ) -> Iterator[GenerateOutput]:
    # prepare prompt
    prompt_tokens = tokenizer.encode(prompt)
    prompt_tokens = remove_bos_duplicates(token_ids=prompt_tokens, bos_token_id=tokenizer.bos_token_id)

    # prepare for generation
    detokenizer = tokenizer._detokenizer
    detokenizer.reset()
    finish_reason = None
    eos_token = tokenizer.eos_token
    stop = [] if stop is None else stop
    if eos_token is not None:
        stop.append(eos_token)
    stop = list(set(stop))
    stop = list(filter(lambda x: x != '', stop)) # remove empty strings
    stop_tuple = [(s, len(s)) for s in stop]
    stop_tuple.sort(key=lambda x: x[1], reverse=True)
    eos_tuple = [eos_token, len(eos_token)] if eos_token else None
    stop_suffix = None
    prompt_len = len(prompt_tokens)
    text = ''
    for step, n in zip(
        generate_step(prompt=mx.array(prompt_tokens), model=model, cache_history=cache_history, verbose=verbose, **kwargs),
        range(max_tokens)
    ):
        if n + 1 == max_tokens:
            finish_reason = 'length'
        detokenizer.add_token(step.token)
        text += detokenizer.last_segment
        sc = stopping_criteria(text=text, stop_tuple=stop_tuple, eos_tuple=eos_tuple)
        if sc.stop_met:
            if sc.trim_length:
                stop_suffix = text[-sc.trim_length:]
                finish_reason = 'stop'
            break
        if any(
            (sequence_overlap(text, s) for s in stop)
        ):
            continue
        new_text = text
        text = ''
        yield GenerateOutput(text=new_text, step=step, finish_reason=finish_reason, prompt_len=prompt_len)

    detokenizer.finalize()
    text += detokenizer.last_segment
    if text:
        if stop_suffix is not None:
            text = text[: -len(stop_suffix)]
        yield GenerateOutput(text=text, step=step, finish_reason=finish_reason, prompt_len=prompt_len)

def generate(
        model: nn.Module, 
        tokenizer: TokenizerWrapper, 
        prompt: str,
        max_tokens: int = 512,
        stop: Optional[List[str]] = None,
        cache_history: Union[StepOutput, CacheHistory] = None,
        verbose: bool = False,
        **kwargs
    ):
    # prepare prompt
    prompt_tokens = tokenizer.encode(prompt)
    prompt_tokens = remove_bos_duplicates(token_ids=prompt_tokens, bos_token_id=tokenizer.bos_token_id)

    # prepare for generation
    detokenizer = tokenizer._detokenizer
    detokenizer.reset()
    finish_reason = None
    eos_token = tokenizer.eos_token
    stop = [] if stop is None else stop
    if eos_token is not None:
        stop.append(eos_token)
    stop = list(set(stop))
    stop = list(filter(lambda x: x != '', stop)) # remove empty strings
    stop_tuple = [(s, len(s)) for s in stop]
    stop_tuple.sort(key=lambda x: x[1], reverse=True)
    eos_tuple = [eos_token, len(eos_token)] if eos_token else None
    stop_suffix = None
    prompt_len = len(prompt_tokens)
    text = ''
    for step, n in zip(
        generate_step(prompt=mx.array(prompt_tokens), model=model, cache_history=cache_history, verbose=verbose, **kwargs),
        range(max_tokens)
    ):
        if n + 1 == max_tokens:
            finish_reason = 'length'
        detokenizer.add_token(step.token)
        text += detokenizer.last_segment
        sc = stopping_criteria(text=text, stop_tuple=stop_tuple, eos_tuple=eos_tuple)
        if sc.stop_met:
            if sc.trim_length:
                stop_suffix = text[-sc.trim_length:]
                finish_reason = 'stop'
            break

    detokenizer.finalize()
    text += detokenizer.last_segment
    if stop_suffix is not None:
        text = text[: -len(stop_suffix)]
    return GenerateOutput(text=text, step=step, finish_reason=finish_reason, prompt_len=prompt_len)



from .utils import CacheHistory, StepOutput, save_cache, load_cache, find_max_prefix_num
import json, os
from logging import Logger
from datetime import datetime as dt
from typing import Optional, List, Union, Dict, NamedTuple, Tuple

class PromptSummary(NamedTuple):
    token_ids: List[int]
    length: int
    last_modified: dt

def prompt_summary_to_dict(ps: PromptSummary) -> Dict[str, Union[List[int], float]]:
    return dict(token_ids=ps.token_ids, last_modified=ps.last_modified.timestamp())

class CacheManager:
    """Class to manage multiple cache histories.
    """
    def __init__(self, cache_dir: str, token_threshold: int = 50, max_keep: int = 10, logger: Optional[Logger] = None) -> None:
        """Initialise the cache manager.

        Args:
            cache_dir (str): The directory to save all the caches.
            token_threshold (int, optional): Minimum number of tokens to be considered to save as a prompt cache history. Defaults to 50.
            max_keep (int, optional): Maximum number of cache history files to keep. Defaults to 10.
            logger (Optional[Logger], optional): If a logger is provided, messages of adding and deleting caches will be added.
        """
        self.cache_dir = os.path.abspath(cache_dir)
        self.json_dir = os.path.join(self.cache_dir, 'cached_prompts.json')
        self.token_threshold = token_threshold
        self.max_keep = max_keep
        self.logger = logger
        if not os.path.exists(self.json_dir):
            os.makedirs(self.cache_dir, exist_ok=True)
            prompts = dict()
            with open(self.json_dir, 'w') as f:
                json.dump(prompts, f)

    @property
    def cached_prompts(self) -> Dict[str, PromptSummary]:
        """Dictionary that contains the information of all the currently stored cache histories.

        Returns:
            Dict[str, PromptSummary]: Dictionary that contains the information of all the currently stored cache histories.
        """
        with open(self.json_dir, 'r') as f:
            prompts = json.load(f)
        for k, v in prompts.items():
            v['last_modified'] = dt.fromtimestamp(v['last_modified'])
            v['length'] = len(v['token_ids'])
            prompts[k] = PromptSummary(**v)   
        return prompts
    
    @property
    def new_cache_id(self) -> str:
        """An unused cache ID for a new cache history.

        Returns:
            str: An unused cache ID for a new cache history.
        """
        current_ids: List[str] = list(self.cached_prompts.keys())
        current_ids = [int(x.removeprefix('cache_')) for x in current_ids]
        if len(current_ids) == 0:
            return 'cache_0'
        max_id = max(current_ids)
        for i in range(max_id):
            if i not in current_ids:
                return f'cache_{i}'
        return f'cache_{max_id + 1}'
    
    def _log_info(self, message: str) -> None:
        if self.logger is not None:
            self.logger.info(message)
    
    def drop_cache(self, cache_id: str) -> None:
        """Drop an existing cache history.

        Args:
            cache_id (str): The cache ID of the file to drop.
        """
        cached_prompts = self.cached_prompts
        if cache_id in cached_prompts.keys():
            ps = cached_prompts.pop(cache_id)
            cache_file = os.path.join(self.cache_dir, f'{cache_id}.safetensors')
            if os.path.exists(cache_file):
                os.remove(cache_file)
                self._log_info(f'Cache history "{cache_id}" dropped. Number of tokens: {ps.length}.')
        for k, v in cached_prompts.items():
            cached_prompts[k] = prompt_summary_to_dict(v)
        with open(self.json_dir, 'w') as f:
            json.dump(cached_prompts, f, indent=4)

    def find_cache(self, token_ids: List[int], token_threshold: Optional[int] = None) -> Tuple[Optional[CacheHistory], Optional[str]]:
        """Find a cache history given the new prompt token ids.

        Args:
            token_ids (List[int]): The new prompt token ids.
            token_threshold (Optional[int], optional): Minimum number of tokens in the cache history to be considered to use. If None is given, token_threshold of the cache manager will be used as default. Defaults to None.

        Returns:
            Tuple[Optional[CacheHistory], Optional[str]]: The cache history and it's cache ID.
        """
        self._log_info(f'Number of tokens for prompt: {len(token_ids)}')
        cached_prompts = list(self.cached_prompts.items())
        if len(cached_prompts) == 0:
            self._log_info('No existing cache.')
            return None, None
        token_threshold = self.token_threshold if token_threshold is None else token_threshold
        cached_prompts = [(x, find_max_prefix_num(new=token_ids, baseline=x[1].token_ids)) for x in cached_prompts]
        cached_prompts = list(filter(lambda x: x[1] >= token_threshold, cached_prompts))
        if len(cached_prompts) == 0:
            self._log_info(f'No cache with prefix match tokens larger than {token_threshold}.')
            return None, None
        cached_prompts.sort(key=lambda x: x[1], reverse=True)
        if cached_prompts[0][1] == 0:
            return None, None
        key = cached_prompts[0][0][0]
        cache_file = os.path.join(self.cache_dir, f'{key}.safetensors')
        cache_history, metadata = load_cache(cache_file)
        self.update_cache_time(cache_id=key)
        self._log_info(f'Cache history "{key}" with {cached_prompts[0][0][1].length} tokens is selected.')
        return cache_history, key
    
    def update_cache_time(self, cache_id: str) -> None:
        """Updating the last modified time of an existing cache history.

        Args:
            cache_id (str): The existing cache ID to update.
        """
        cached_prompts = self.cached_prompts
        for k, v in cached_prompts.items():
            cached_prompts[k] = prompt_summary_to_dict(v)
            if k == cache_id:
                cached_prompts[k]['last_modified'] = dt.now().timestamp()
        with open(self.json_dir, 'w') as f:
            json.dump(cached_prompts, f, indent=4)

    def save_cache(self, 
        cache: Union[StepOutput, CacheHistory],
        replace_threshold: float = 0.9,
        token_threshold: Optional[int] = None, 
        from_cache_id: Optional[str] = None, 
        metadata: Optional[Dict[str, str]] = None
    ) -> None:
        """Save a new cache history.

        Args:
            cache (Union[StepOutput, CacheHistory]): The new cache history to save.
            replace_threshold (float, optional): Percentage threshold of tokens prefix similar to the given from_cache_id to replace the given cache ID. Defaults to 0.9.
            token_threshold (Optional[int], optional): Minimum number of tokens in the cache history to be considered to be saved. If None is given, token_threshold of the cache manager will be used as default. Defaults to None.
            from_cache_id (Optional[str], optional): A cache ID that will be considered to be replaced (see replace_threshold). Defaults to None.
            metadata (Optional[Dict[str, str]], optional): Metadata to save in the cache history file. Defaults to None.
        """
        token_threshold = self.token_threshold if token_threshold is None else token_threshold
        new_len = len(cache.token_ids)
        if new_len < token_threshold:
            return
        cached_prompts = self.cached_prompts
        if from_cache_id in cached_prompts.keys():
            old_len = cached_prompts[from_cache_id].length
            max_prefix = find_max_prefix_num(new=cache.token_ids, baseline=cached_prompts[from_cache_id].token_ids)
            similarity = max_prefix / old_len
            if similarity > replace_threshold:
                if old_len > new_len:
                    self.update_cache_time(cache_id=from_cache_id)
                    return
                else:
                    self.drop_cache(cache_id=from_cache_id)
                    cached_prompts.pop(from_cache_id)
        
        duplicates = list(filter(lambda x: x[1].length == find_max_prefix_num(cache.token_ids, x[1].token_ids), cached_prompts.items()))
        for k, v in duplicates:
            self.drop_cache(cache_id=k)
            cached_prompts.pop(k)
                
        if len(cached_prompts) >= self.max_keep:
            exist_caches = list(cached_prompts.keys())
            cached_prompts = list(cached_prompts.items())
            cached_prompts.sort(key=lambda x: x[1].last_modified, reverse=True)
            cached_prompts = dict(cached_prompts[:self.max_keep - 1])
            new_caches = cached_prompts.keys()
            for cache_id in exist_caches:
                if cache_id not in new_caches:
                    self.drop_cache(cache_id=cache_id)

        new_id = self.new_cache_id
        ps = PromptSummary(token_ids=cache.token_ids, length=new_len, last_modified=dt.now())
        cached_prompts[new_id] = ps
        save_cache(cache=cache, filename=os.path.join(self.cache_dir, f'{new_id}.safetensors'), metadata=metadata)
        for k, v in cached_prompts.items():
            cached_prompts[k] = prompt_summary_to_dict(v)
        with open(self.json_dir, 'w') as f:
            json.dump(cached_prompts, f, indent=4)
        self._log_info(f'New cache history "{new_id}" saved. Number of tokens: {new_len}.')
        return



from .engine import ModelConfig
from .utils import PACKAGE_NAME
import yaml
from argparse import ArgumentParser

parser = ArgumentParser(prog=f'{PACKAGE_NAME}.create_config_file',
    description='Creating a model configuration file for model serving.')

parser.add_argument('-n', '--num-models', type=int, default=1, help='Number of model examples in the config file.')

def main():
    args = parser.parse_args()
    num_models = args.num_models
    configs = [ModelConfig(model_id_or_path=f'/path/to/model_{i}')._asdict() for i in range(num_models)]
    with open('model_configs.yaml', 'w') as f:
        f.write(yaml.dump(configs, sort_keys=False))

if __name__ == '__main__':
    main()
